@@ -5,17 +5,26 @@ Changes to the content of one single LDAP entry.
 changing of location in tree)
 """
 
-from twisted.python.util import InsensitiveDict
 from ldaptor import attributeset
-from ldaptor.protocols import pureldap
+from ldaptor.protocols import pureldap, pureber
 from ldaptor.protocols.ldap import ldif, distinguishedname
 
 class Modification(attributeset.LDAPAttributeSet):
     def patch(self, entry):
         raise NotImplementedError
 
+    _LDAP_OP = None
+
     def asLDAP(self):
-        raise NotImplementedError
+        if self._LDAP_OP is None:
+            raise NotImplementedError("%s.asLDAP not implemented"
+                                      % self.__class__.__name__)
+        return str(pureber.BERSequence([
+            pureber.BEREnumerated(self._LDAP_OP),
+            pureber.BERSequence([ pureldap.LDAPAttributeDescription(self.key),
+                                  pureber.BERSet(map(pureldap.LDAPString, list(self))),
+                                  ]),
+            ]))
 
     def __eq__(self, other):
         if not isinstance(other, self.__class__):
@@ -23,15 +32,14 @@ class Modification(attributeset.LDAPAttributeSet):
         return super(Modification, self).__eq__(other)
 
 class Add(Modification):
+    _LDAP_OP = 0
+
     def patch(self, entry):
         if self.key in entry:
             entry[self.key].update(self)
         else:
             entry[self.key] = self
 
-    def asLDAP(self):
-	return pureldap.LDAPModification_add(self.key, list(self))
-
     def asLDIF(self):
         r=[]
         values = list(self)
@@ -43,6 +51,8 @@ class Add(Modification):
         return ''.join(r)
 
 class Delete(Modification):
+    _LDAP_OP = 1
+
     def patch(self, entry):
         if not self:
             del entry[self.key]
@@ -50,9 +60,6 @@ class Delete(Modification):
             for v in self:
                 entry[self.key].remove(v)
 
-    def asLDAP(self):
-	return pureldap.LDAPModification_delete(self.key, list(self))
-
     def asLDIF(self):
         r=[]
         values = list(self)
@@ -64,6 +71,8 @@ class Delete(Modification):
         return ''.join(r)
 
 class Replace(Modification):
+    _LDAP_OP = 2
+
     def patch(self, entry):
         if self:
             entry[self.key] = self
@@ -73,9 +82,6 @@ class Replace(Modification):
             except KeyError:
                 pass
 
-    def asLDAP(self):
-	return pureldap.LDAPModification_replace(self.key, list(self))
-
     def asLDIF(self):
         r=[]
         values = list(self)
@@ -115,11 +121,45 @@ class ModifyOp(Operation):
         r.append("\n")
         return ''.join(r)
 
+    def asLDAP(self):
+        return pureldap.LDAPModifyRequest(
+            object=str(self.dn),
+            modification=[x.asLDAP() for x in self.modifications])
+
+    def _getClassFromOp(class_, op):
+        for mod in [Add, Delete, Replace]:
+            if op == mod._LDAP_OP:
+                return mod
+        return None
+    _getClassFromOp = classmethod(_getClassFromOp)
+
+    def fromLDAP(class_, request):
+        if not isinstance(request, pureldap.LDAPModifyRequest):
+            raise RuntimeError("%s.fromLDAP needs an LDAPModifyRequest"
+                               % class_.__name__)
+        dn = request.object
+        result = []
+        for op, mods in request.modification:
+            op = op.value
+            klass = class_._getClassFromOp(op)
+            if klass is None:
+                raise RuntimeError("Unknown LDAP op number %r in %s.fromLDAP"
+                                   % (op, class_.__name__))
+
+            key, vals = mods
+            key = key.value
+            vals = [x.value for x in vals]
+            m = klass(key, vals)
+            result.append(m)
+        return class_(dn, result)
+    fromLDAP = classmethod(fromLDAP)
+
     def patch(self, root):
         d = root.lookup(self.dn)
         def gotEntry(entry, modifications):
             for mod in self.modifications:
                 mod.patch(entry)
+            return entry
         d.addCallback(gotEntry, self.modifications)
         return d
 
@@ -132,16 +172,16 @@ class ModifyOp(Operation):
                 + ')')
 
     def __eq__(self, other):
-	if not isinstance(other, self.__class__):
-	    return 0
-	if self.dn != other.dn:
-	    return 0
+        if not isinstance(other, self.__class__):
+            return 0
+        if self.dn != other.dn:
+            return 0
         if self.modifications != other.modifications:
             return 0
         return 1
 
     def __ne__(self, other):
-	return not self==other
+        return not self==other
 
 class AddOp(Operation):
     def __init__(self, entry):
@@ -167,14 +207,14 @@ class AddOp(Operation):
                 + ')')
 
     def __eq__(self, other):
-	if not isinstance(other, self.__class__):
-	    return False
-	if self.entry != other.entry:
-	    return False
+        if not isinstance(other, self.__class__):
+            return False
+        if self.entry != other.entry:
+            return False
         return True
 
     def __ne__(self, other):
-	return not self==other
+        return not self==other
 
 class DeleteOp(Operation):
     def __init__(self, dn):
@@ -201,12 +241,12 @@ class DeleteOp(Operation):
                 + ')')
 
     def __eq__(self, other):
-	if not isinstance(other, self.__class__):
-	    return False
-	if self.dn != other.dn:
-	    return False
+        if not isinstance(other, self.__class__):
+            return False
+        if self.dn != other.dn:
+            return False
         return True
 
     def __ne__(self, other):
-	return not self==other
+        return not self==other
 
